@@ -19,19 +19,16 @@ import os
 import datetime
 import http.client
 from urllib.parse import urlparse
+from typing import List, Iterator, IO
 
 import boto3
 import botocore
 import botocore.exceptions
-
-from typing import List, Iterator, BinaryIO
-
 from google.cloud import storage
 from google.protobuf import timestamp_pb2, duration_pb2
 
-from nsl.stac import gcs_storage_client, bearer_auth
-
-from nsl.stac import StacItem, Asset, TimestampField, Eo, DatetimeRange
+from nsl.stac import gcs_storage_client, bearer_auth, \
+    StacItem, Asset, TimestampField, Eo, DatetimeRange
 from nsl.stac.enum import Band, CloudPlatform, FieldRelationship, SortDirection, AssetType
 
 DEFAULT_RGB = [Band.RED, Band.GREEN, Band.BLUE, Band.NIR]
@@ -54,7 +51,7 @@ def _gcp_blob_metadata(bucket: str, blob_name: str) -> storage.Blob:
 
 def download_gcs_object(bucket: str,
                         blob_name: str,
-                        file_obj: BinaryIO = None,
+                        file_obj: IO[bytes] = None,
                         save_filename: str = "",
                         make_dir=True) -> str:
     """
@@ -75,7 +72,7 @@ def download_gcs_object(bucket: str,
 
     if file_obj is not None:
         blob.download_to_file(file_obj=file_obj, client=gcs_storage_client)
-        if "name" in file_obj:
+        if "name" in file_obj.__dict__:
             save_filename = file_obj.name
         else:
             save_filename = ""
@@ -92,7 +89,7 @@ def download_gcs_object(bucket: str,
 
 def download_s3_object(bucket: str,
                        blob_name: str,
-                       file_obj: BinaryIO = None,
+                       file_obj: IO = None,
                        save_filename: str = ""):
     # TODO, can this be global?
     s3 = boto3.resource('s3')
@@ -100,7 +97,7 @@ def download_s3_object(bucket: str,
         bucket_obj = s3.Bucket(bucket)
         if file_obj is not None:
             bucket_obj.download_fileobj(blob_name, file_obj)
-            if "name" in file_obj:
+            if "name" in file_obj.__dict__:
                 save_filename = file_obj.name
             else:
                 save_filename = ""
@@ -119,7 +116,7 @@ def download_s3_object(bucket: str,
             raise
 
 
-def download_href_object(asset: Asset, file_obj: BinaryIO = None, save_filename: str = ""):
+def download_href_object(asset: Asset, file_obj: IO = None, save_filename: str = ""):
     """
     download the href of an asset
     :param asset: The asset to download
@@ -133,6 +130,9 @@ def download_href_object(asset: Asset, file_obj: BinaryIO = None, save_filename:
     if len(asset.type) > 0:
         headers["content-type"] = asset.type
 
+    if not asset.href:
+        raise ValueError("no href on asset")
+
     host = urlparse(asset.href)
     asset_url = "/download/{object}".format(object=asset.object_path)
     conn = http.client.HTTPConnection(host.netloc)
@@ -140,14 +140,14 @@ def download_href_object(asset: Asset, file_obj: BinaryIO = None, save_filename:
 
     res = conn.getresponse()
     if res.status is not 200:
-        raise ValueError("{path} does not exist".format(path=asset_url))
+        raise ValueError("{path} does not exist".format(path=asset.href))
 
     if len(save_filename) > 0:
         with open(save_filename, mode='wb') as f:
             f.write(res.read())
     elif file_obj is not None:
         file_obj.write(res.read())
-        if "name" in file_obj:
+        if "name" in file_obj.__dict__:
             save_filename = file_obj.name
         else:
             save_filename = ""
@@ -160,7 +160,7 @@ def download_href_object(asset: Asset, file_obj: BinaryIO = None, save_filename:
 
 def download_asset(asset: Asset,
                    from_bucket: bool = False,
-                   file_obj: BinaryIO = None,
+                   file_obj: IO = None,
                    save_filename: str = "",
                    save_directory: str = ""):
     """
