@@ -22,7 +22,7 @@ import io
 from google.protobuf import timestamp_pb2
 from datetime import datetime, timezone, date, timedelta
 
-from nsl.stac import StacRequest, LandsatRequest, MosaicRequest
+from nsl.stac import StacRequest, LandsatRequest, MosaicRequest, EoRequest
 from nsl.stac import StacItem, Asset, TimestampField, GeometryData, SpatialReferenceData, Mosaic
 from nsl.stac import utils, enum
 from nsl.stac.enum import AssetType, Band, CloudPlatform, Constellation, FieldRelationship
@@ -275,6 +275,20 @@ class TestLandsat(unittest.TestCase):
         number = client.count(stac_request)
         self.assertEquals(1, number)
 
+    def test_2000(self):
+        start = datetime(1999, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
+        end = datetime(1999, 4, 6, 12, 52, 59, tzinfo=timezone.utc)
+        observed_range = utils.pb_timestampfield(rel_type=FieldRelationship.BETWEEN, start=start, end=end)
+
+        stac_request = StacRequest(observed=observed_range, limit=20, landsat=LandsatRequest())
+        for stac_item in client.search(stac_request):
+            self.assertEquals(Constellation.LANDSAT, stac_item.eo.constellation)
+            print(datetime.fromtimestamp(stac_item.datetime.seconds, tz=timezone.utc))
+            self.assertGreaterEqual(utils.pb_timestamp(end).seconds, stac_item.datetime.seconds)
+            self.assertLessEqual(utils.pb_timestamp(start).seconds, stac_item.datetime.seconds)
+
+        self.assertEquals(2728, client.count(stac_request))
+
     def test_count_more(self):
         start = datetime(2014, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
         end = datetime(2014, 4, 1, 12, 52, 59, tzinfo=timezone.utc)
@@ -293,15 +307,44 @@ class TestLandsat(unittest.TestCase):
 
 
 class TestDatetimeQueries(unittest.TestCase):
-    def test_date_GT_OR_EQ(self):
-        bd = date(2015, 11, 3)
-        observed_range = TimestampField(value=utils.pb_timestamp(bd),
-                                        rel_type=FieldRelationship.GT_OR_EQ)
-        stac_request = StacRequest(observed=observed_range)
+    def test_date_LT_OR_EQ(self):
+        bd = date(2014, 11, 3)
+        observed_range = utils.pb_timestampfield(rel_type=FieldRelationship.LT_OR_EQ, value=bd)
+        eo = EoRequest(constellation=enum.Constellation.NAIP)
+        stac_request = StacRequest(observed=observed_range, eo=eo)
         stac_item = client.search_one(stac_request)
         self.assertIsNotNone(stac_item)
         self.assertLessEqual(utils.pb_timestamp(bd).seconds, stac_item.datetime.seconds)
 
+    def test_date_GT_OR_EQ(self):
+        bd = date(2015, 11, 3)
+        observed_range = TimestampField(value=utils.pb_timestamp(bd, tzinfo=timezone.utc),
+                                        rel_type=FieldRelationship.GT_OR_EQ)
+        stac_request = StacRequest(observed=observed_range)
+        stac_item = client.search_one(stac_request)
+        self.assertIsNotNone(stac_item)
+        self.assertLessEqual(utils.pb_timestamp(bd, tzinfo=timezone.utc).seconds, stac_item.observed.seconds)
+
+    @unittest.skip("mono-241 publishing failure")
+    def test_date_GT_OR_EQ_datetime(self):
+        bd = date(2015, 11, 3)
+        observed_range = TimestampField(value=utils.pb_timestamp(bd, tzinfo=timezone.utc),
+                                        rel_type=FieldRelationship.GT_OR_EQ)
+        stac_request = StacRequest(observed=observed_range)
+        stac_item = client.search_one(stac_request)
+        self.assertIsNotNone(stac_item)
+        self.assertLessEqual(utils.pb_timestamp(bd, tzinfo=timezone.utc).seconds, stac_item.datetime.seconds)
+
+    def test_observed_GT(self):
+        bdt = datetime(2015, 11, 3, 1, 1, 1, tzinfo=timezone.utc)
+        observed_range = TimestampField(value=utils.pb_timestamp(bdt),
+                                        rel_type=FieldRelationship.GT)
+        stac_request = StacRequest(observed=observed_range)
+        stac_item = client.search_one(stac_request)
+        self.assertIsNotNone(stac_item)
+        self.assertLessEqual(utils.pb_timestamp(bdt).seconds, stac_item.observed.seconds)
+
+    @unittest.skip("mono-241 publishing failure")
     def test_datetime_GT(self):
         bdt = datetime(2015, 11, 3, 1, 1, 1, tzinfo=timezone.utc)
         observed_range = TimestampField(value=utils.pb_timestamp(bdt),
@@ -351,6 +394,7 @@ class TestDatetimeQueries(unittest.TestCase):
                             utils.pb_timestamp(start).seconds < stac_item.datetime.seconds)
         self.assertEqual(count, 5)
 
+    @unittest.skip("mono-241 publishing failure")
     def test_datetime_not_range_desc(self):
         start = datetime(2013, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
         end = datetime(2014, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
@@ -366,6 +410,21 @@ class TestDatetimeQueries(unittest.TestCase):
             self.assertTrue(utils.pb_timestamp(end).seconds < stac_item.datetime.seconds)
         self.assertEqual(count, 5)
 
+    def test_observed_not_range_desc(self):
+        start = datetime(2013, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
+        end = datetime(2014, 4, 1, 12, 45, 59, tzinfo=timezone.utc)
+        observed_range = TimestampField(start=utils.pb_timestamp(start),
+                                        stop=utils.pb_timestamp(end),
+                                        rel_type=FieldRelationship.NOT_BETWEEN,
+                                        sort_direction=enum.SortDirection.DESC)
+        stac_request = StacRequest(observed=observed_range, limit=5)
+        count = 0
+        for stac_item in client.search(stac_request):
+            count += 1
+            print(datetime.fromtimestamp(stac_item.observed.seconds, tz=timezone.utc))
+            self.assertTrue(utils.pb_timestamp(end).seconds < stac_item.observed.seconds)
+        self.assertEqual(count, 5)
+
     def test_date_utc_eq(self):
         value = date(2019, 8, 6)
         texas_utc_offset = timezone(timedelta(hours=-6))
@@ -465,8 +524,21 @@ class TestHelpers(unittest.TestCase):
                 data3 = f_obj.read().decode('ascii')
                 self.assertMultiLineEqual(data1, data3)
 
+    def test_download_geotiff(self):
+        import os
+        stac_request = StacRequest(id='20190822T183518Z_746_POM1_ST2_P')
+
+        stac_item = client.search_one(stac_request)
+
+        # get the Geotiff asset from the assets map
+        asset = utils.get_asset(stac_item, asset_type=enum.AssetType.GEOTIFF)
+
+        with tempfile.TemporaryDirectory() as d:
+            file_path = utils.download_asset(asset=asset, save_directory=d)
+            print("{0} has {1} bytes".format(os.path.basename(file_path), os.path.getsize(file_path)))
+
     def test_download_href(self):
-        stac_id = "20190829T172947Z_1619_POM1_ST2_P"
+        stac_id = "20190829T173549Z_1799_POM1_ST2_P"
         stac_item = client.search_one(stac_request=StacRequest(id=stac_id))
         asset = utils.get_asset(stac_item, asset_type=AssetType.THUMBNAIL)
 
